#ifndef ADAPTIVE_SAMPLES_H
#define ADAPTIVE_SAMPLES_H


#include <algorithm>
#include <fstream>

#include "misc/utils.h"


namespace CSA
{

//--------------------------------------------------------------------------

/*
  This uses both POSITIONS and RANGES parts of external module interface.
  FIXME Currently supports only one sequence.
*/

template<class Index>
class AdaptiveSamples
{
  public:
    typedef uint key_type;
    typedef std::pair<key_type, key_type> sample_type;

    AdaptiveSamples(const Index& _index, const std::string& base_name) :
      index(_index), samples(0), inverse_samples(0), text_start(0),
      items(0), inverse_items(0),
      ok(false)
    {
      std::string file_name = base_name + SA_SAMPLES_EXTENSION;
      std::ifstream sample_file(file_name.c_str(), std::ios_base::binary);
      if(!sample_file)
      {
        std::cerr << "AdaptiveSamples: Cannot open sample file!" << std::endl;
        return;
      }
      sample_file.read((char*)&(this->size), sizeof(this->size)); // Actually text size.
      sample_file.read((char*)&(this->size), sizeof(this->size));

      this->samples = new sample_type[this->size];
      this->inverse_samples = new sample_type[this->size];
      for(key_type i = 0; i < this->size; i++)
      {
        this->samples[i] = sample_type(this->index.getSize(), this->index.getSize());
        this->inverse_samples[i] = this->samples[i];
      }

      pair_type buffer;
      for(usint i = 0; i < this->size; i++)
      {
        sample_file.read((char*)&buffer, sizeof(pair_type));
        this->addSample(sample_type(buffer.first, buffer.second));
        this->addInverseSample(sample_type(buffer.second, buffer.first));
        if(buffer.second == 0) { this->text_start = buffer.first; }
      }

      sample_file.close();
      this->ok = true;
    }

    ~AdaptiveSamples()
    {
      delete[] this->samples; this->samples = 0;
      delete[] this->inverse_samples; this->inverse_samples = 0;
    }

    void writeTo(std::ofstream& sample_file) const
    {
      pair_type buffer(this->index.getSize(), this->items);
      sample_file.write((char*)&buffer, sizeof(pair_type));

      for(key_type i = 0; i < this->size; i++)
      {
        if(this->samples[i].first < this->index.getSize())
        {
          buffer = pair_type(this->samples[i].first, this->samples[i].second);
          sample_file.write((char*)&buffer, sizeof(pair_type));
        }
      }
    }

    bool isOk() const { return this->ok; }

//--------------------------------------------------------------------------

    usint locate(usint i, bool steps)
    {
      if(i >= this->index.getSize()) { return (steps ? 0 : this->index.getSize()); }

      usint offset = 0;
      sample_type sample(i, 0);
      this->index.convertToBWTIndex(i);

      while(true)
      {
        if(this->index.hasImplicitSample(i))
        {
          sample.second = this->index.getImplicitSample(i) - offset;
          break;
        }
        this->index.convertToSAIndex(i);
        key_type j = this->hash(i);
        if(this->samples[j].first == i)
        {
          sample.second = this->samples[j].second - offset;
          break;
        }
        i = this->index.psi(i); offset++;
      }

      this->addSample(sample);
      return (steps ? offset : sample.second);
    }

    usint* locate(pair_type range, bool steps)
    {
      if(length(range) == 0) { return 0; }

      usint* data = new usint[length(range)];
      for(usint i = 0, j = range.first; j <= range.second; i++, j++)
      {
        data[i] = this->locate(j, steps);
      }

      return data;
    }

    // Returns the number of steps required to find a sample.
    usint display(pair_type range, uchar* data)
    {
      if(isEmpty(range) || range.second >= this->index.getSize() || data == 0) { return 0; }

      usint i = range.first, sa_index = this->index.getSize();
      do
      {
        key_type j = this->hash(i);
        if(this->inverse_samples[j].first == i)
        {
          sa_index = this->inverse_samples[j].second;
          break;
        }
        i--;
      }
      while(i > 0);
      if(sa_index == this->index.getSize()) { sa_index = this->text_start; }
      usint steps = range.first - i;

      for(; i < range.first; i++)
      {
        sa_index = this->index.psi(sa_index);
        this->index.convertToSAIndex(sa_index);
      }
      for(; i <= range.second; i++)
      {
        data[i - range.first] = this->index.getCharacter(sa_index);
        sa_index = this->index.psi(sa_index);
        this->index.convertToSAIndex(sa_index);
      }

      return steps;
    }

//--------------------------------------------------------------------------

    inline bool supportsLocate() const { return true; }
    inline bool supportsDisplay() const { return true; }
    inline usint getNumberOfSamples() const { return this->items; }
    inline usint getNumberOfInverseSamples() const { return this->inverse_items; }
    inline double getLoad() const { return this->items / (double)(this->size); }
    inline double getInverseLoad() const { return this->inverse_items / (double)(this->size); }

    usint reportSize() const { return sizeof(*this) + 2 * this->size * sizeof(sample_type); }

    void report() const
    {
      std::cout << "Adaptive samples: " << std::endl;
      std::cout << "Size:  " << this->reportSize() / (double)MEGABYTE << " MB" << std::endl;
      std::cout << "Load:  " << this->getLoad() << " (regular), "
                             << this->getInverseLoad() << " (inverse)" << std::endl;
      std::cout << std::endl;
    }

//--------------------------------------------------------------------------

  private:
    const Index& index;

    sample_type* samples;
    sample_type* inverse_samples;
    key_type text_start;
    usint size, items, inverse_items;
    bool ok;
    
    void addSample(sample_type sample)
    {
      key_type i = this->hash(sample.first);
      if(this->samples[i].first >= this->index.getSize()) { this->items++; }
      this->samples[i] = sample;
    }

    void addInverseSample(sample_type sample)
    {
      key_type i = this->hash(sample.first);
      if(this->inverse_samples[i].first >= this->index.getSize()) { this->inverse_items++; }
      this->inverse_samples[i] = sample;
    }

    key_type hash(key_type key)
    {
      key_type res = 0;
      uchar* data = (uchar*)&key;
      for(usint i = 0; i < sizeof(key_type); i++) { res = 101 * res + data[i]; }
      return res % this->size;
    }

    // These are not allowed.
    AdaptiveSamples();
    AdaptiveSamples(const AdaptiveSamples&);
    AdaptiveSamples& operator = (const AdaptiveSamples&);
};


//--------------------------------------------------------------------------

};  // namespace CSA


#endif // ADAPTIVE_SAMPLES_H
